from django.apps import apps
from django.contrib.admin.views.decorators import staff_member_required
from django.http import (
    HttpResponse,
    HttpResponseBadRequest,
    HttpResponseNotAllowed,
    JsonResponse,
)
from django.shortcuts import redirect
from django.urls import reverse
from coffee.contrib.parse.functions import get_pagination

from coffee.contrib.decorators import has_sufficient_params
from coffee.contrib.render.functions import (
    render_model_form,
    render_model_table,
)


@staff_member_required
@has_sufficient_params
def get_model_list(request, app_name=None, model_name=None, pk=None):
    page_size = int(request.GET.get("page_size", 15))
    page = int(request.GET.get("page", 1))
    json = request.GET.get("json", None)
    pagination = request.GET.get("pagination", None)

    offset = (page - 1) * page_size

    cls = apps.get_model(app_name, model_name)
    queryset = cls.objects.all()[offset : offset + page_size]

    pagination_json = get_pagination(request, cls, page_size, page)

    html = render_model_table(
        cls, queryset, None if not pagination else pagination_json
    )

    if json:
        return JsonResponse({"html": html, "pagination": pagination_json})

    return HttpResponse(html)


@staff_member_required
@has_sufficient_params
def get_model_form(request, app_name=None, model_name=None, pk=None):
    instance = None

    json = request.GET.get("json", None)

    cls = None
    if pk:
        try:
            cls = apps.get_model(app_name, model_name)
            instance = cls.objects.get(pk=pk)
        except cls.DoesNotExist:
            exception = f"pk '{pk}' for '{app_name}.{model_name}' not found"
            return HttpResponseBadRequest(exception)

    html = render_model_form(request, app_name, model_name, instance)

    if json:
        return JsonResponse({"html": html})

    return HttpResponse(html)


@staff_member_required
@has_sufficient_params
def post_model_form(request, app_name=None, model_name=None, pk=None):
    cls = apps.get_model(app_name, model_name)
    pk_field = cls._meta.pk.name

    if request.method != "POST":
        return HttpResponseNotAllowed(
            f"'{request.method} is not valid for this endpoint."
        )

    data = {
        k: None if v == "" else v
        for k, v in request.POST.copy().items()
        if k != "csrfmiddlewaretoken"
    }
    pk = data.get(pk_field)

    if not pk:
        instance, _ = cls.objects.get_or_create(**data)
        pk = instance.pk

    if pk:
        instance = cls.objects.get(pk=pk)
        for key, value in data.items():
            setattr(instance, key, value)
        instance.save()

    base_url = reverse("coffee_form")
    url = base_url + f"?app_name={app_name}&model_name={model_name}&pk={pk}"

    return redirect(url)


@staff_member_required
@has_sufficient_params
def delete_model_instance(request, app_name=None, model_name=None, pk=None):
    cls = apps.get_model(app_name, model_name)

    if request.method != "POST":
        return HttpResponseNotAllowed(
            f"'{request.method} is not valid for this endpoint."
        )

    if pk:
        instance = cls.objects.get(pk=pk)
        instance.delete()

    base_url = reverse("coffee_form")
    url = base_url + f"?app_name={app_name}&model_name={model_name}"

    return redirect(url)



import math
import re


def get_pagination(request, cls, page_size, page):
    rx = re.compile(r"&page=[0-9]")
    path = request.get_full_path()

    if rx.search(path):
        previous = re.sub(rx, f"&page={page - 1}", path)
        next = re.sub(rx, f"&page={page + 1}", path)
    else:
        previous = path + f"&page={page - 1}"
        next = path + f"&page={page + 1}"

    count = cls.objects.count()
    pages = math.ceil(count / page_size)

    if page - 1 == 0:
        previous = None

    if page + 1 > pages:
        next = None

    return {
        "previous": previous,
        "next": next,
        "current_page": page,
        "number_of_pages": pages,
        "count": count,
    }



def has_sufficient_params(func):
    def wrapper(request, *args, **kwargs):
        app_name = request.GET.get("app_name")
        model_name = request.GET.get("model_name")
        pk = request.GET.get("pk")

        assert app_name is not None, "'app_name' not provided"
        assert model_name is not None, "'model_name' not provided"

        return func(request, app_name, model_name, pk, *args, **kwargs)

    return wrapper



from re import template


class InputTemplate:
    type: str
    tag: str
    implemented: bool
    disabled: bool

    def __init__(self, type="text", tag="input", implemented=True, disabled=False):
        self.tag = tag
        self.type = type
        self.implemented = implemented
        self.disabled = disabled


class HTMLComponent:
    tag = None
    value = None
    children = []
    properties = {}
    class_name = None

    def __init__(self, value=None, children=None, class_name=None):
        self.value = value
        if children:
            self.value = "".join([str(i) for i in children])

        if class_name:
            self.class_name = class_name

    def get_template(self):
        self.properties["class"] = self.class_name
        props = " ".join([f"{k}='{v}'" for k, v in self.properties.items() if v])
        return f"<{self.tag} {props}>{self.value or None}</{self.tag}>"

    def __str__(self):
        return self.get_template()


class Div(HTMLComponent):
    tag = "div"


class InputLabel(HTMLComponent):
    tag = "label"
    field_name: str = None
    class_name = "coffee-form-item-label"

    def get_template(self):
        return f"<label class={self.class_name} for='{self.value}'>{self.value}</label>"


class Input(HTMLComponent):
    template: InputTemplate
    field_name: str
    class_name = "coffee-form-item-input"

    def __init__(self, template, field_name, value=None):
        self.field_name = field_name
        self.value = value
        self.tag = template.tag
        self.type = template.type

    def get_template(self):
        value = self.value or ""
        if self.type == "datetime-local" and self.value:
            value = self.value.strftime("%Y-%m-%dT%H:%M")

        self.properties = {
            "value": value or "",
            "name": self.field_name,
            "id": self.field_name,
            "type": self.type,
            "class": self.class_name,
        }
        props = " ".join([f"{k}='{v}'" for k, v in self.properties.items() if v])

        label_html = str(InputLabel(value=self.field_name))
        if self.type == "hidden" or self.tag == "button":
            label_html = ""

        input_html = f"<{self.tag} {props}></{self.tag}>"
        if self.tag == "textarea" or self.tag == "button":
            input_html = f"<{self.tag} {props}>{value}</{self.tag}>"

        html = str(
            Div(children=[label_html, input_html], class_name="coffee-form-item")
        )

        return html


class SubmitButton(HTMLComponent):
    class_name = "coffee-form-submit"

    def get_template(self):
        return f'<button type="submit" class="{self.class_name}">{self.value}</button>'


class THead(HTMLComponent):
    tag = "thead"
    children = []
    class_name = "coffee-table-thead"


class TBody(HTMLComponent):
    tag = "tbody"
    children = []
    class_name = "coffee-table-tbody"

    def __init__(self, children):
        if children:
            self.value = "".join([str(i) for i in children])


class Td(HTMLComponent):
    tag = "td"
    class_name = "coffee-table-td"


class TRow(HTMLComponent):
    tag = "tr"
    class_name = "coffee-table-tr"


class Table(HTMLComponent):
    tag = "table"
    thead: THead
    tbody: TBody
    class_name = "coffee-table"

    def __init__(self, thead=None, tbody=None):
        self.thead = str(thead)
        self.tbody = str(tbody)
        self.value = self.thead + self.tbody

    def get_template(self):
        return super().get_template()

    def __str__(self):
        return self.get_template()


class Form(HTMLComponent):
    action: str
    children: list[Input]
    tag = "form"
    class_name = "coffee-form"

    def __init__(self, children, action, method="POST"):
        if children:
            self.value = "".join([str(i) for i in children])

        self.properties = {
            "action": action,
            "method": method,
        }


class PaginationButton(HTMLComponent):
    class_name = "coffee-pagination-button"
    tag = "a"
    href: str

    def __init__(self, href=None, value=None):
        self.value = value
        self.properties = {"href": href}

        if not href:
            self.properties["disabled"] = "true"


class PaginationText(HTMLComponent):
    class_name = "coffee-pagination-text"
    tag = "span"


class Pagination(HTMLComponent):
    class_name = "coffee-pagination"
    tag = "div"

    def __init__(self, pagination):
        previous = pagination.get("previous")
        next = pagination.get("next")
        number_of_pages = pagination.get("number_of_pages")
        current_page = pagination.get("current_page")

        children = [
            PaginationButton(href=previous, value="previous"),
            PaginationText(value=f"{current_page} / {number_of_pages}"),
            PaginationButton(href=next, value="next"),
        ]
        if children:
            self.value = "".join([str(i) for i in children])



import re


def to_camel_case(text):
    first, *others = text.split("_")

    return "".join([first.lower(), *map(str.title, others)])


def to_snake_case(text):
    text = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", text)
    text = re.sub("__([A-Z])", r"_\1", text)
    text = re.sub("([a-z0-9])([A-Z])", r"\1_\2", text)

    return text.lower()



import re
from importlib import import_module
from inspect import getmembers, isclass

from django.urls import reverse

from coffee.exceptions import DefinitionNotFound, FieldNotSupported
from coffee.contrib.render.components import (
    Form,
    Input,
    InputTemplate,
    Pagination,
    SubmitButton,
    TBody,
    THead,
    TRow,
    Table,
    Td,
)
from coffee.contrib.render.definitions import (
    FORM_TRANSLATION,
)
from django.db.models.base import ModelBase
from django.forms.models import model_to_dict
from django.middleware.csrf import get_token


def get_model_definitions(app_name):
    module = import_module(f"{app_name}.models")
    models = getmembers(module, predicate=isclass)

    definitions = {}

    for model_name, model in models:
        if type(model) == ModelBase and hasattr(model, "_meta"):
            rx = re.compile(r"(?:<class [\"'])(.+)(?:[\"']>)")
            match = rx.findall(repr(model))
            match = match[0].split(".")[0]

            if match == app_name:
                if model_name not in definitions:
                    definitions[model_name] = {}

                for field in model._meta.fields:
                    related_model = None

                    if field.related_model:
                        related_model = field.related_model.__name__

                    definitions[model_name][field.name] = {
                        "related_model": related_model,
                        "primary_key": field.primary_key,
                        "db_type": field.get_internal_type(),
                        "nullable": field.null,
                    }

    return definitions


def get_input_template(field):
    db_type = field.get("db_type")
    if db_type not in FORM_TRANSLATION:
        raise FieldNotSupported()
    template = FORM_TRANSLATION[db_type]

    return template


def get_csrf_token(request):
    csrf_token = get_token(request)
    template = InputTemplate(type="hidden")
    return Input(template=template, field_name="csrfmiddlewaretoken", value=csrf_token)


def render_model_form(request, app_name, model_name, instance):
    definitions = get_model_definitions(app_name)

    if model_name not in definitions:
        raise DefinitionNotFound(model_name)

    definition = definitions[model_name]

    base_url = reverse("coffee_post")

    action = base_url + f"?app_name={app_name}&model_name={model_name}"

    children = []
    for name, field in definition.items():
        template = get_input_template(field)
        if template.implemented is True:
            value = getattr(instance, name, None)
            item = Input(field_name=name, template=template, value=value)
            children.append(item)

    children.append(get_csrf_token(request))

    button_text = "Submit"
    if instance:
        button_text = "Update"

    children.append(SubmitButton(value=button_text))

    form = Form(children=children, action=action)

    delete_button = ""
    if instance:
        delete_button = get_delete_button_html(request, app_name, model_name, instance)

    return str(form) + str(delete_button)


def render_model_table(cls, queryset, pagination):
    keys = [i.name for i in cls._meta.fields]
    headers = [Td(i) for i in keys]
    thead = THead(value=TRow(children=headers))

    trs = []
    for row in queryset:
        row_dict = model_to_dict(row)
        tds = []
        for key in keys:
            tds.append(Td(value=row_dict.get(key)))
        trs.append(TRow(children=tds))

    tbody = TBody(children=trs)

    html = str(Table(tbody=tbody, thead=thead))

    if pagination:
        pagination_html = Pagination(pagination=pagination)
        html += str(pagination_html)

    return html


def get_delete_button_html(request, app_name, model_name, pk):
    base_url = reverse("coffee_delete")
    action = base_url + f"?app_name={app_name}&model_name={model_name}&pk={pk}"
    children = [get_csrf_token(request), SubmitButton(value="Delete")]
    form = Form(action=action, children=children)
    return form



from django.urls import path

from coffee.views import (
    delete_model_instance,
    get_model_form,
    get_model_list,
    post_model_form,
)


urlpatterns = [
    path("form/", get_model_form, name="coffee_form"),
    path("form/submit/", post_model_form, name="coffee_post"),
    path("delete/", delete_model_instance, name="coffee_delete"),
    path("list/", get_model_list, name="coffee_list"),
]



from coffee.contrib.render import InputTemplate


TYPESCRIPT_TRANSLATION = {
    "AutoField": "number",
    "BigAutoField": "number",
    "BigIntegerField": "number",
    "BinaryField": "number",
    "BooleanField": "boolean",
    "CharField": "text",
    "CommaSeparatedIntegerField": "number",
    "DateField": "Date",
    "DateTimeField": "Date",
    "DecimalField": "number",
    "DurationField": "number",
    "EmailField": "text",
    "FilePathField": "text",
    "FloatField": "number",
    "GenericIPAddressField": "text",
    "IPAddressField": "text",
    "IntegerField": "number",
    "NullBooleanField": "boolean | null",
    "PositiveBigIntegerField": "number",
    "PositiveIntegerField": "number",
    "PositiveSmallIntegerField": "number",
    "SlugField": "string",
    "SmallAutoField": "number",
    "SmallIntegerField": "number",
    "TextField": "string",
    "TimeField": "string",
    "URLField": "string",
    "UUIDField": "string",
    "ForeignKey": "number",
    "OneToOneField": "number",
}

FORM_TRANSLATION = {
    "AutoField": InputTemplate(type="number", tag="input"),
    "BigAutoField": InputTemplate(type="number", tag="input"),
    "BigIntegerField": InputTemplate(type="number", tag="input"),
    "BinaryField": InputTemplate(type="number", tag="input"),
    "BooleanField": InputTemplate(type="checkbox", tag="input"),
    "CharField": InputTemplate(type="text", tag="input"),
    "CommaSeparatedIntegerField": InputTemplate(type="number", tag="input"),
    "DateField": InputTemplate(type="date", tag="input"),
    "DateTimeField": InputTemplate(type="datetime-local", tag="input"),
    "DecimalField": InputTemplate(type="number", tag="input"),
    "DurationField": InputTemplate(type="number", tag="input"),
    "EmailField": InputTemplate(type="text", tag="input"),
    "FilePathField": InputTemplate(type="text", tag="input"),
    "FloatField": InputTemplate(type="number", tag="input"),
    "GenericIPAddressField": InputTemplate(type="text", tag="input"),
    "IPAddressField": InputTemplate(type="text", tag="input"),
    "IntegerField": InputTemplate(type="number", tag="input"),
    "NullBooleanField": InputTemplate(type="checkbox", tag="input"),
    "PositiveBigIntegerField": InputTemplate(type="number", tag="input"),
    "PositiveIntegerField": InputTemplate(type="number", tag="input"),
    "PositiveSmallIntegerField": InputTemplate(type="number", tag="input"),
    "SlugField": InputTemplate(type="text", tag="input"),
    "SmallAutoField": InputTemplate(type="number", tag="input"),
    "SmallIntegerField": InputTemplate(type="number", tag="input"),
    "TextField": InputTemplate(type="text", tag="textarea"),
    "TimeField": InputTemplate(type="text", tag="input"),
    "URLField": InputTemplate(type="text", tag="input"),
    "UUIDField": InputTemplate(type="text", tag="input"),
    "ForeignKey": InputTemplate(type="number", tag="input"),
    "OneToOneField": InputTemplate(type="number", tag="input"),
    "SearchVectorField": InputTemplate(type="text", tag="textarea", implemented=False),
    "GeometryField": InputTemplate(type="text", tag="textarea", implemented=False),
}



from coffee.exceptions import FieldNotSupported
from coffee.contrib.render.definitions import TYPESCRIPT_TRANSLATION


def get_typescript_type(db_type):
    if db_type not in TYPESCRIPT_TRANSLATION:
        raise FieldNotSupported(db_type)

    return TYPESCRIPT_TRANSLATION[db_type]
